STRIPE_GENERIC = "STRIPE_50000"
STRIPE_INVALID_REQUEST = "STRIPE_40002"
STRIPE_BAD_REQUEST = "STRIPE_40003"

# Customer
STRIPE_CREATE_CUSTOMER_METHOD = "CREATE_CUSTOMER"
STRIPE_MODIFY_CUSTOMER_METHOD = "MODIFY_CUSTOMER"
STRIPE_DELETE_CUSTOMER_METHOD = "DELETE_CUSTOMER"
GENERATE_FULL_NAME = "GENERATE_FULL_NAME"
GENERATE_CUSTOMER_ADDRESS = "GENERATE_CUSTOMER_ADDRESS"

# Product
STRIPE_CREATE_PRODUCT_METHOD = "CREATE_PRODUCT"
STRIPE_MODIFY_PRODUCT_METHOD = "MODIFY_PRODUCT"
STRIPE_RETRIEVE_PRODUCT_METHOD = "RETRIEVE_PRODUCT"
STRIPE_DELETE_PRODUCT_METHOD = "DELETE_PRODUCT"
GET_PRICE_DATA = "GET_PRICE_DATA"

# Price
STRIPE_CREATE_PRICE_METHOD = "CREATE_PRICE"
STRIPE_MODIFY_PRICE_METHOD = "MODIFY_PRICE"
STRIPE_RETRIEVE_PRICE_METHOD = "RETRIEVE_PRICE"

# messages
STRIPE_INVALID_REQUEST_MESSAGE = "Stripe invalid request error."
STRIPE_ERROR_MESSAGE = "There is some issue while processing request with stripe."

# Subscription
STRIPE_CREATE_SUBSCRIPTION_METHOD = "CREATE_SUBSCRIPTION"
STRIPE_MODIFY_SUBSCRIPTION_METHOD = "MODIFY_SUBSCRIPTION"
STRIPE_CANCEL_SUBSCRIPTION_METHOD = "CANCEL_SUBSCRIPTION"
STRIPE_RETRIEVE_SUBSCRIPTION_METHOD = "RETRIEVE_SUBSCRIPTION"

# Invoice
STRIPE_CREATE_INVOICE_METHOD = "CREATE_INVOICE"
STRIPE_FINALIZE_INVOICE_METHOD = "FINALIZE_INVOICE"
STRIPE_RETRIEVE_INVOICE_METHOD = "RETRIEVE_INVOICE"

# Invoice-item
STRIPE_CREATE_INVOICE_ITEM_METHOD = "CREATE_INVOICE_ITEM"

# Subscription Schedule
STRIPE_CREATE_SUBSCRIPTION_SCHEDULE_METHOD = "CREATE_SUBSCRIPTION_SCHEDULE"
STRIPE_MODIFY_SUBSCRIPTION_SCHEDULE_METHOD = "MODIFY_SUBSCRIPTION_SCHEDULE"
STRIPE_RELEASE_SUBSCRIPTION_SCHEDULE_METHOD = "RELEASE_SUBSCRIPTION_SCHEDULE"

# Coupon
STRIPE_CREATE_COUPON_METHOD = "CREATE_COUPON"
STRIPE_MODIFY_COUPON_METHOD = "MODIFY_COUPON"
STRIPE_RETRIEVE_COUPON_METHOD = "RETRIEVE_COUPON"
STRIPE_DELETE_COUPON_METHOD = "DELETE_COUPON"

# Promo-code
STRIPE_CREATE_PROMOCODE_METHOD = "CREATE_PROMOCODE"
STRIPE_LIST_PROMOCODE_METHOD = "LIST_PROMOCODE"
STRIPE_RETRIEVE_PROMOCODE_METHOD = "RETRIEVE_COUPON"



import stripe
from src.common.constants.constant import *
from src.exceptions.exception_constants import *
from src.exceptions.exception_handling import *


class StripeSubscriptionSchedule:

    @staticmethod
    def create_stripe_subscription_schedule(
            stripe_secret_key: str, subscription_id: str
    ) -> dict:
        """
        A stripe create subscription schedule method.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account’s API keys. Stripe raises an invalid request
                                  error if you don’t include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type subscription_id: string
        :param subscription_id: Stripe subscription id.

        :type: dictionary
        :returns: Returns a subscription schedule object if the call succeeded.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(subscription_id, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_CREATE_SUBSCRIPTION_SCHEDULE_METHOD)

            subscription_schedule = stripe.SubscriptionSchedule.create(
                from_subscription=subscription_id,
                api_key=stripe_secret_key
            )

            return {
                DATA: subscription_schedule,
                MESSAGE: "Subscription schedule created successfully."
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_CREATE_SUBSCRIPTION_SCHEDULE_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_CREATE_SUBSCRIPTION_SCHEDULE_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_CREATE_SUBSCRIPTION_SCHEDULE_METHOD)

    @staticmethod
    def update_stripe_subscription_schedule(
            stripe_secret_key: str, schedule_id: str,
            current_phase_start_dt: str, current_phase_end_dt: str,
            current_price_id: str, new_price_id: str = None,
            coupon: str = None, default_payment_method: str = None,
            metadata: dict = None, description: str = None
    ) -> dict:
        """
        A stripe update subscription schedule method.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account’s API keys. Stripe raises an invalid request
                                  error if you don’t include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type schedule_id: string
        :param schedule_id: Stripe subscription schedule id.

        :type current_phase_start_dt: string
        :param current_phase_start_dt: The date at which this phase of the subscription
                                       schedule starts or now. Must be set on the first phase.

        :type current_phase_end_dt: string
        :param current_phase_end_dt: The date at which this phase of the subscription
                                     schedule ends. If set, iterations must not be set.

        :type current_price_id: string
        :param current_price_id: Current ID of the price object.

        :type new_price_id: string
        :param new_price_id: New ID of the price object.

        :type coupon: string
        :param coupon: The identifier of the coupon to apply to this phase of the subscription
                       schedule.

        :type default_payment_method: string
        :param default_payment_method: ID of the default payment method for the subscription
                                       schedule. It must belong to the customer associated with
                                       the subscription schedule. If not set, invoices will use
                                       the default payment method in the customer’s invoice
                                       settings.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to a phase. Metadata on a
                         schedule’s phase will update the underlying subscription’s metadata
                         when the phase is entered, adding new keys and replacing existing keys
                         in the subscription’s metadata. Individual keys in the subscription’s
                         metadata can be unset by posting an empty value to them in the phase’s
                         metadata. To unset all keys in the subscription’s metadata, update the
                         subscription directly or unset every key individually from the phase’s
                         metadata.

        :type description: string
        :param description: Subscription description, meant to be displayable to the customer.
                            Use this field to optionally store an explanation of the subscription.

        :type: dictionary
        :returns: Returns an updated subscription schedule object if the call succeeded.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(schedule_id, str) \
                    or not isinstance(current_phase_start_dt, int) \
                    or not isinstance(current_phase_end_dt, int) \
                    or not isinstance(current_price_id, str) \
                    or not isinstance(new_price_id, (str, type(None))) \
                    or not isinstance(coupon, (str, type(None))) \
                    or not isinstance(default_payment_method, (str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(description, (str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_MODIFY_SUBSCRIPTION_SCHEDULE_METHOD)

            if new_price_id is None:
                new_price_id = current_price_id

            subscription_schedule = stripe.SubscriptionSchedule.modify(
                schedule_id,
                api_key=stripe_secret_key,
                end_behavior="release",
                phases=[
                    {
                        "start_date": current_phase_start_dt,
                        "end_date": current_phase_end_dt,
                        "items": [
                            {
                                "price": current_price_id,
                            },
                        ],
                    },
                    {
                        "items": [
                            {
                                "price": new_price_id,
                            },
                        ],
                        "metadata": metadata,
                        "coupon": coupon,
                        "default_payment_method": default_payment_method,
                        "description": description,
                    },
                ],
            )

            return {
                DATA: subscription_schedule,
                MESSAGE: "Subscription Schedule modified successfully."
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_MODIFY_SUBSCRIPTION_SCHEDULE_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_MODIFY_SUBSCRIPTION_SCHEDULE_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_MODIFY_SUBSCRIPTION_SCHEDULE_METHOD)

    @staticmethod
    def release_stripe_subscription_schedule(
            stripe_secret_key: str, schedule_id: str
    ) -> dict:
        """
        A stripe release subscription schedule method.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account’s API keys. Stripe raises an invalid request
                                  error if you don’t include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type schedule_id: string
        :param schedule_id: Stripe schedule id.

        :type: dictionary
        :returns: The released subscription_schedule object. Its status will be released,
                  released_at will be the current time, and released_subscription will be
                  the ID of the subscription schedule managed prior to being released.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(schedule_id, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_RELEASE_SUBSCRIPTION_SCHEDULE_METHOD)

            subscription_schedule = stripe.SubscriptionSchedule.release(
                    schedule_id, api_key=stripe_secret_key
                )

            return {
                DATA: subscription_schedule,
                MESSAGE: "Subscription schedule released successfully."
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_RELEASE_SUBSCRIPTION_SCHEDULE_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_RELEASE_SUBSCRIPTION_SCHEDULE_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_RELEASE_SUBSCRIPTION_SCHEDULE_METHOD)



import stripe
from src.common.constants.constant import *
from src.common.constants.product import *
from src.common.service import convert_to_cents
from src.exceptions.exception_constants import *

from src.exceptions.exception_handling import *




class StripePrice:

    @staticmethod
    def create_stripe_price(
            stripe_secret_key: str,
            product_id: str,
            currency: str,
            unit_amount: float,
            active: bool = True,
            recurring_interval: str = None,
            recurring_interval_count: int = None,
            metadata: dict = None,
            nickname: str = None) -> dict:
        """
        Create new stripe price for stripe product.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type product_id: string
        :param product_id: The ID of the product that this price will belong to.

        :type currency: string
        :param currency: Three-letter ISO currency code, in lowercase.
                         Must be a supported currency.

        :type unit_amount: float
        :param unit_amount: :type price_unit_amount: float
        :param price_unit_amount: A positive value or 0 for a free price (rupees, usd, 
                                  etc) representing how much to charge. More than 2 
                                  decimal is rounded to 2 decimal places.

        :type active: boolean
        :param active: Whether the price can be used for new purchases. Defaults to true.

        :type recurring_interval: string
        :param recurring_interval: Specifies billing frequency. 
                                   Either day, week, month or year.

        :type recurring_interval_count: integer
        :param recurring_interval_count: The number of intervals between subscription billings.
                                         For example, interval=month and interval_count=3 bills
                                         every 3 months. Maximum of one-year interval allowed
                                         (1 year, 12 months, or 52 weeks).

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can 
                         be useful for storing additional information about the object in a 
                         structured format. Individual keys can be unset by posting an empty 
                         value to them. All keys can be unset by posting an empty value to 
                         metadata.

        :type nickname: string
        :param nickname: A brief description of the price, hidden from customers.

        :type: dictionary
        :return: The newly created Price object is returned upon success.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(currency, str) \
                    or not isinstance(product_id, str) \
                    or not isinstance(unit_amount, (int, float)) \
                    or not isinstance(recurring_interval, (str, type(None))) \
                    or not isinstance(recurring_interval_count, (int, str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_CREATE_PRICE_METHOD)

            recurring = None
            if recurring_interval:
                recurring = {
                    INTERVAL: recurring_interval,
                    INTERVAL_COUNT: recurring_interval_count
                }

            price = stripe.Price.create(
                api_key=stripe_secret_key,
                currency=currency,
                product=product_id,
                unit_amount=convert_to_cents(unit_amount),
                active=active,
                recurring=recurring,
                nickname=nickname,
                metadata=metadata
            )

            return {
                MESSAGE: "Stripe price created successfully.",
                DATA: price
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_CREATE_PRICE_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code=STRIPE_GENERIC,
                                      method_name=STRIPE_CREATE_PRICE_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_CREATE_PRICE_METHOD)

    @staticmethod
    def modify_stripe_price(
            stripe_secret_key: str,
            price_id: str,
            active: bool = True,
            metadata: dict = None,
            nickname: str = None) -> dict:
        """
        Update a stripe price.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type price_id: string
        :param price_id: An ID of price that needs to be updated.

        :type active: boolean
        :param active: Whether the price can be used for new purchases. Defaults to true.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can 
                         be useful for storing additional information about the object in a 
                         structured format. Individual keys can be unset by posting an empty 
                         value to them. All keys can be unset by posting an empty value to 
                         metadata.

        :type nickname: string
        :param nickname: A brief description of the price, hidden from customers.

        :type: dictionary
        :return: The updated price object is returned upon success.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(price_id, str) \
                    or not isinstance(active, bool) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(nickname, (str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_MODIFY_PRICE_METHOD)

            price = stripe.Price.modify(
                sid=price_id,
                api_key=stripe_secret_key,
                active=active,
                metadata=metadata,
                nickname=nickname
            )

            return {
                MESSAGE: "Stripe price modified successfully.",
                DATA: price
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_MODIFY_PRICE_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code=STRIPE_GENERIC,
                                      method_name=STRIPE_MODIFY_PRICE_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_MODIFY_PRICE_METHOD)

    @staticmethod
    def retrieve_stripe_price(
            stripe_secret_key: str,
            price_id: str) -> dict:
        """
        retrieve price information from price id.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type price_id: string
        :param price_id: An ID of price that needs to be fetched.

        :type: dictionary
        :return: Returns a price if a valid price or plan ID was provided.
        """
        try:
            if not isinstance(price_id, str) or not isinstance(stripe_secret_key, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_RETRIEVE_PRICE_METHOD)

            price = stripe.Price.retrieve(id=price_id, api_key=stripe_secret_key)

            return {
                MESSAGE: "Stripe price retrieved successfully.",
                DATA: price
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_RETRIEVE_PRICE_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code=STRIPE_GENERIC,
                                      method_name=STRIPE_RETRIEVE_PRICE_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_RETRIEVE_PRICE_METHOD)



import stripe

from src.common.constants.constant import BAD_REQUEST_MESSAGE, DATA, MESSAGE
from src.common.constants.product import CURRENCY, INTERVAL, INTERVAL_COUNT, RECURRING, UNIT_AMOUNT
from src.common.service import convert_to_cents
from src.exceptions.exception_constants import GET_PRICE_DATA, STRIPE_BAD_REQUEST, \
    STRIPE_CREATE_PRODUCT_METHOD, STRIPE_DELETE_PRODUCT_METHOD, STRIPE_GENERIC, STRIPE_INVALID_REQUEST, \
    STRIPE_INVALID_REQUEST_MESSAGE, STRIPE_MODIFY_PRODUCT_METHOD, STRIPE_RETRIEVE_PRODUCT_METHOD
from src.exceptions.exception_handling import CoreBadRequest, CoreGenericException, CoreStripeException


def get_price_data(
    currency: str,
    unit_amount: float,
    recurring_interval: str,
    recurring_interval_count: str):
    try:
        if currency:
            if not unit_amount:
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=GET_PRICE_DATA)

            return {
                CURRENCY: currency,
                UNIT_AMOUNT: convert_to_cents(unit_amount),
                RECURRING: {
                    INTERVAL: recurring_interval,
                    INTERVAL_COUNT: recurring_interval_count
                }
            }
        else:
            return None

    except CoreBadRequest as exception:
        raise CoreBadRequest(exception_detail=exception.exception_detail,
                                exception_code=exception.exception_code,
                                method_name=exception.method_name)

    except Exception as exception:
        raise CoreGenericException(exception_traceback=exception,
                                    exception_code=STRIPE_GENERIC,
                                    method_name=GET_PRICE_DATA)


class StripeProduct:

    @staticmethod
    def create_stripe_product(
            stripe_secret_key: str,
            name: str,
            description: str = None,
            price_currency: str = None,
            price_unit_amount: float = None,
            price_recurring_interval: str = None,
            price_recurring_interval_count: int = None,
            metadata: dict = None) -> dict:
        """
        Creating a new stripe product.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type name: string
        :param name: The product's name, meant to be displayable to the customer.

        :type description: string
        :param description: The product's description, meant to be displayable to the customer. 
                            Use this field to optionally store a long form explanation of the 
                            product being sold for your own rendering purposes.

        :type price_currency: string
        :param price_currency: Three-letter ISO currency code, in lowercase. 
                               Must be a supported currency.

        :type price_unit_amount: float
        :param price_unit_amount: A positive value or 0 for a free price (rupees, usd, 
                                  etc) representing how much to charge. More than 2 
                                  decimal is rounded to 2 decimal places.

        :type price_recurring_interval: string
        :type price_recurring_interval: Specifies billing frequency. 
                                        Either day, week, month or year.

        :type price_recurring_interval_count: integer
        :param price_recurring_interval_count: The number of intervals between subscription 
                                               billings. For example, interval=month and 
                                               interval_count=3 bills every 3 months. Maximum 
                                               of one-year interval allowed (1 year, 12 months, 
                                               or 52 weeks).

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can 
                         be useful for storing additional information about the object in a 
                         structured format. Individual keys can be unset by posting an empty 
                         value to them. All keys can be unset by posting an empty value to 
                         metadata.

        :type: dictionary
        :return: returns a product object if the call succeeded.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(name, str) \
                    or not isinstance(description, (str, type(None))) \
                    or not isinstance(price_currency, (str, type(None))) \
                    or not isinstance(price_unit_amount, (int, float, type(None))) \
                    or not isinstance(price_recurring_interval, (str, type(None))) \
                    or not isinstance(price_recurring_interval_count, (int, str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_CREATE_PRODUCT_METHOD)

            product = stripe.Product.create(
                api_key=stripe_secret_key,
                name=name,
                description=description,
                metadata=metadata,
                default_price_data=get_price_data(
                    price_currency,
                    price_unit_amount,
                    price_recurring_interval,
                    price_recurring_interval_count
                )
            )

            return {
                MESSAGE: "Stripe product created successfully.",
                DATA: product
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_traceback=exception, 
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 method_name=STRIPE_CREATE_PRODUCT_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code = STRIPE_GENERIC,
                                      method_name=STRIPE_CREATE_PRODUCT_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception, 
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_CREATE_PRODUCT_METHOD)

    @staticmethod
    def modify_stripe_product(
            stripe_secret_key: str,
            product_id: str,
            name: str = None,
            description: str = None,
            default_price_id: str = None,
            active: bool = True,
            metadata: dict = None) -> dict:
        """
        Modify stripe product details.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type product_id: string
        :param product_id: stripe product id which needs to be updated.

        :type name: string
        :param name: The product's name, meant to be displayable to the customer.

        :type description: string
        :param description: The product's description, meant to be displayable to the customer. 
                            Use this field to optionally store a long form explanation of the 
                            product being sold for your own rendering purposes.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can 
                         be useful for storing additional information about the object in a 
                         structured format. Individual keys can be unset by posting an empty 
                         value to them. All keys can be unset by posting an empty value to 
                         metadata.

        :type default_price_id: string
        :param default_price_id: The ID of the <a href="https://stripe.com/docs/api/prices">
                                 Price</a> object that is the default price for this product. 
                                 Only those IDs who are attached to product will be able to 
                                 set as default.

        :type active: boolean
        :param active: Whether the product is available for purchase.

        :type: dictionary
        :returns: Returns the product object if the update succeeded.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(product_id, str) \
                    or not isinstance(description, (str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(default_price_id, (str, type(None))) \
                    or not isinstance(name, (str, type(None))) \
                    or not isinstance(active, bool):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_MODIFY_PRODUCT_METHOD)

            product = stripe.Product.modify(
                sid=product_id,
                api_key=stripe_secret_key,
                name=name,
                description=description,
                metadata=metadata,
                default_price=default_price_id,
                active=active
            )

            return {
                MESSAGE: "Stripe product modified successfully.",
                DATA: product
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_traceback=exception, 
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 method_name=STRIPE_MODIFY_PRODUCT_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code = STRIPE_GENERIC,
                                      method_name=STRIPE_MODIFY_PRODUCT_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception, 
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_MODIFY_PRODUCT_METHOD)

    @staticmethod
    def retrieve_stripe_product(
        stripe_secret_key: str, 
        product_id: str) -> dict:
        """
        Retrieve stripe product details using product id.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type product_id: string
        :param product_id: Unique product ID from either a product creation request or 
                           the product list, and Stripe will return the corresponding 
                           product information.

        :type: dictionary
        :return: Returns a product object if a valid identifier was provided.
        """
        try:
            if not isinstance(stripe_secret_key, str) or not isinstance(product_id, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_RETRIEVE_PRODUCT_METHOD)

            product = stripe.Product.retrieve(id=product_id, api_key=stripe_secret_key)

            return {
                MESSAGE: "Stripe product retrieved successfully.",
                DATA: product
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_traceback=exception, 
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 method_name=STRIPE_RETRIEVE_PRODUCT_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                       exception_code = STRIPE_GENERIC,
                                       method_name=STRIPE_RETRIEVE_PRODUCT_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception, 
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_RETRIEVE_PRODUCT_METHOD)

    @staticmethod
    def delete_stripe_product(stripe_secret_key: str, product_id: str) -> dict:
        """
        Delete a stripe product using product id.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type product_id: string
        :param product_id: Unique product ID which needs to be deleted.

        :type: dictionary
        :return: Returns a deleted object on success. Otherwise, this call raises an error.
        """
        try:
            if not isinstance(product_id, str) or not isinstance(stripe_secret_key, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_DELETE_PRODUCT_METHOD)

            product = stripe.Product.delete(sid=product_id, api_key=stripe_secret_key)

            return {
                MESSAGE: "Stripe product deleted successfully.",
                DATA: product
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_traceback=exception, 
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 method_name=STRIPE_DELETE_PRODUCT_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                       exception_code = STRIPE_GENERIC,
                                       method_name=STRIPE_DELETE_PRODUCT_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception, 
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_DELETE_PRODUCT_METHOD)



import stripe
from src.common.constants.constant import BAD_REQUEST_MESSAGE, DATA, MESSAGE
from src.common.constants.customer import CITY, COUNTRY, DEFAULT_PAYMENT_METHOD, \
    LINE1, LINE2, POSTAL_CODE, STATE
from src.exceptions.exception_constants import GENERATE_CUSTOMER_ADDRESS, \
    GENERATE_FULL_NAME, STRIPE_BAD_REQUEST, STRIPE_CREATE_CUSTOMER_METHOD, \
    STRIPE_DELETE_CUSTOMER_METHOD, STRIPE_GENERIC, STRIPE_INVALID_REQUEST, \
    STRIPE_INVALID_REQUEST_MESSAGE, STRIPE_MODIFY_CUSTOMER_METHOD
from src.exceptions.exception_handling import CoreBadRequest, CoreGenericException, CoreStripeException


def generate_full_name(first_name: str = None, last_name: str = None):
    try:
        if first_name and last_name:
            return "{0} {1}".format(first_name.strip(), last_name.strip())
        elif first_name:
            return first_name.strip()
        elif last_name:
            return last_name.strip()
        else:
            return None

    except Exception as exception:
        raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                    exception_traceback=exception,
                                    method_name=GENERATE_FULL_NAME)

def generate_customer_address(
        city: str = None,
        country: str = None,
        address_line1: str = None,
        address_line2: str = None,
        postal_code: str = None,
        state: str = None):
    try:
        if (address_line1 is None or address_line1 == "") \
                and (address_line2 is None or address_line2 == "") \
                and (postal_code is None or postal_code == "") \
                and (state is None or state == "") \
                and (country is None or country == "") \
                and (city is None or city == ""):
            return None

        return {
            CITY: city,
            COUNTRY: country,
            LINE1: address_line1,
            LINE2: address_line2,
            POSTAL_CODE: postal_code,
            STATE: state
        }

    except Exception as exception:
        raise CoreGenericException(exception_traceback=exception,
                                    exception_code=STRIPE_GENERIC,
                                    method_name=GENERATE_CUSTOMER_ADDRESS)


class StripeCustomer:

    @staticmethod
    def create_stripe_customer(
            stripe_secret_key: str,
            email: str,
            first_name: str,
            last_name: str = None,
            address_line1: str = None,
            address_line2: str = None,
            city: str = None,
            postal_code: str = None,
            state: str = None,
            country: str = None,
            description: str = None,
            phone: str = None,
            metadata: dict = None) -> dict:
        """
        Create new stripe Customer.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type email: string
        :param email: Customer's email address.

        :type first_name: string
        :param first_name: Customer's valid first name (name).

        :type last_name: string
        :param last_name: Customer's last name (surname). If last_name is not provided 
                          then first name is used for the stripe customer name.

        :type address_line1: string
        :param address_line1: Address line 1 (e.g., street, PO Box, or company name).

        :type address_line2: string
        :param address_line2: Address line 2 (e.g., apartment, suite, unit, or building).

        :type city: string
        :param city: City, district, suburb, town, or village.

        :type country: string
        :param country: Two-letter country code
                        <a href="https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">
                        (ISO 3166-1 alpha-2)</a>.

        :type postal_code: string
        :param postal_code: ZIP or postal code.

        :type state: string
        :param state: State, county, province, or region.

        :type description: string
        :param description: An arbitrary string attached to the object. 
                            Often useful for displaying to users.

        :type phone: string
        :param phone: phone number with country code.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can 
                         be useful for storing additional information about the object in a 
                         structured format. Individual keys can be unset by posting an empty 
                         value to them. All keys can be unset by posting an empty value to 
                         metadata.

        :type: dictionary
        :returns: Returns the customer object if the update succeeded with message.
                  If create parameters are invalid errors will be raised.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(email, str) \
                    or not (isinstance(first_name, str) and first_name.strip() != "") \
                    or not isinstance(last_name, (str, type(None))) \
                    or not isinstance(description, (str, type(None))) \
                    or not isinstance(phone, (str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(address_line1, (str, type(None))) \
                    or not isinstance(address_line2, (str, type(None))) \
                    or not isinstance(city, (str, type(None))) \
                    or not isinstance(country, (str, type(None))) \
                    or not isinstance(postal_code, (str, type(None))) \
                    or not isinstance(state, (str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_CREATE_CUSTOMER_METHOD)

            address = generate_customer_address(
                city=city,
                country=country,
                address_line1=address_line1,
                address_line2=address_line2,
                postal_code=postal_code,
                state=state
            )
            customer = stripe.Customer.create(
                api_key=stripe_secret_key,
                email=email,
                name=generate_full_name(first_name, last_name),
                address=address,
                description=description,
                phone=phone,
                metadata=metadata
            )

            return {
                MESSAGE: "Stripe customer created successfully.",
                DATA: customer
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_CREATE_CUSTOMER_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code=STRIPE_GENERIC,
                                      method_name=STRIPE_CREATE_CUSTOMER_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_CREATE_CUSTOMER_METHOD)


    @staticmethod
    def modify_stripe_customer(
            stripe_secret_key: str,
            customer_id: str,
            email: str = None,
            first_name: str = None,
            last_name: str = None,
            address_line1: str = None,
            address_line2: str = None,
            city: str = None,
            postal_code: str = None,
            state: str = None,
            country: str = None,
            description: str = None,
            phone: str = None,
            default_payment_method: str = None,
            metadata: dict = None) -> dict:
        """
        Update stripe Customer details.

        For modifying customers address user should pass all address parameters 
        (address_line1, address_line2, city, state, country, postal_code) for 
        proper modification. If any parameter doesn't passed, then method will 
        take default value and set it to None (or null). If all parameters are 
        passed as null string or None, then address dictionary will be None and 
        it will not update any address information.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type customer_id: string
        :param customer_id: Stripe customer id for modifying object.

        :type email: string
        :param email: Customer's email address.

        :type first_name: string
        :param first_name: Customer's valid first (name) name.

        :type last_name: string
        :param last_name: Customer's last name (surname).

        :type address_line1: string
        :param address_line1: Address line 1 (e.g., street, PO Box, or company name).

        :type address_line2: string
        :param address_line2: Address line 2 (e.g., apartment, suite, unit, or building).

        :type city: string
        :param city: City, district, suburb, town, or village.

        :type country: string
        :param country: Two-letter country code
                        <a href="https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">
                        (ISO 3166-1 alpha-2)</a>.

        :type postal_code: string
        :param postal_code: ZIP or postal code.

        :type state: string
        :param state: State, county, province, or region.

        :type description: string
        :param description: An arbitrary string attached to the object. 
                            Often useful for displaying to users.

        :type phone: string
        :param phone: phone number with country code.

        :type default_payment_method: string
        :param default_payment_method: Add payment_method id to this field for future payment.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can 
                         be useful for storing additional information about the object in a 
                         structured format. Individual keys can be unset by posting an empty 
                         value to them. All keys can be unset by posting an empty value to 
                         metadata.

        :type: dictionary
        :returns: Returns the updated customer object if the update succeeded.
                  If create parameters are invalid errors will be raised.
        """
        try:
            if not isinstance(customer_id, str) \
                    or not isinstance(stripe_secret_key, str) \
                    or not isinstance(email, (str, type(None))) \
                    or not isinstance(first_name, (str, type(None))) \
                    or not isinstance(last_name, (str, type(None))) \
                    or not isinstance(description, (str, type(None))) \
                    or not isinstance(phone, (str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(address_line1, (str, type(None))) \
                    or not isinstance(address_line2, (str, type(None))) \
                    or not isinstance(city, (str, type(None))) \
                    or not isinstance(country, (str, type(None))) \
                    or not isinstance(postal_code, (str, type(None))) \
                    or not isinstance(state, (str, type(None))) \
                    or not isinstance(default_payment_method, (str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_MODIFY_CUSTOMER_METHOD)

            address = generate_customer_address(
                city=city,
                state=state,
                country=country,
                address_line1=address_line1,
                address_line2=address_line2,
                postal_code=postal_code
            )
            modified_customer = stripe.Customer.modify(
                api_key=stripe_secret_key,
                sid=customer_id,
                email=email,
                name=generate_full_name(first_name, last_name),
                address=address,
                description=description,
                phone=phone,
                invoice_settings={
                    DEFAULT_PAYMENT_METHOD: default_payment_method,
                },
                metadata=metadata
            )

            return {
                MESSAGE: "Stripe customer modified successfully.",
                DATA: modified_customer
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_traceback=exception,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 method_name=STRIPE_MODIFY_CUSTOMER_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_traceback=exception,
                                      exception_code=STRIPE_GENERIC,
                                      method_name=STRIPE_MODIFY_CUSTOMER_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_BAD_REQUEST,
                                       method_name=STRIPE_MODIFY_CUSTOMER_METHOD)

    @staticmethod
    def delete_stripe_customer(stripe_secret_key: str, customer_id: str) -> dict:
        """
        Delete stripe customer using customer id.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account's API keys. Stripe raises an invalid request
                                  error if you don't include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type customer_id: string
        :param customer_id: Stripe customer ID which needs to be deleted permanently.

        :type: dictionary
        :return: returns object of deleted customer.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(customer_id, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_DELETE_CUSTOMER_METHOD)

            deleted_customer = stripe.Customer.delete(
                sid=customer_id,
                api_key=stripe_secret_key
            )

            return {
                MESSAGE: "Stripe customer deleted successfully.",
                DATA: deleted_customer
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreGenericException(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                       exception_traceback=exception,
                                       exception_code=STRIPE_INVALID_REQUEST,
                                       method_name=STRIPE_DELETE_CUSTOMER_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_DELETE_CUSTOMER_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_traceback=exception,
                                       exception_code=STRIPE_GENERIC,
                                       method_name=STRIPE_DELETE_CUSTOMER_METHOD)



from http import HTTPStatus


class CoreGenericException(Exception):
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR

    def __init__(self, exception_code=None, method_name=None,
                 exception_detail="There is some issue. Please contact support.",
                 exception_traceback=None):
        super().__init__(exception_detail, exception_code)
        self.method_name = method_name
        self.exception_code = exception_code
        self.exception_detail = exception_detail
        self.exception_traceback = exception_traceback


class CoreBadRequest(Exception):
    status_code = HTTPStatus.BAD_REQUEST

    def __init__(self, exception_code=None, method_name=None,
                 exception_detail="Bad Request", exception_traceback=None):
        super().__init__(exception_detail, exception_code)
        self.method_name = method_name
        self.exception_code = exception_code
        self.exception_detail = exception_detail
        self.exception_traceback = exception_traceback


class CoreNotFound(Exception):
    status_code = HTTPStatus.NOT_FOUND

    def __init__(self, exception_code=None, method_name=None,
                 exception_detail="Not Found", exception_traceback=None):
        super().__init__(exception_detail, exception_code)
        self.method_name = method_name
        self.exception_code = exception_code
        self.exception_detail = exception_detail
        self.exception_traceback = exception_traceback


class CoreAuthenticationFailed(Exception):
    status_code = HTTPStatus.UNAUTHORIZED

    def __init__(self, exception_code=None, method_name=None,
                 exception_detail="Authentication Failed", exception_traceback=None):
        super().__init__(exception_detail, exception_code)
        self.method_name = method_name
        self.exception_code = exception_code
        self.exception_detail = exception_detail
        self.exception_traceback = exception_traceback


class CorePermissionDenied(Exception):
    status_code = HTTPStatus.FORBIDDEN

    def __init__(self, exception_code=None, method_name=None,
                 exception_detail="Permission Denied", exception_traceback=None):
        super().__init__(exception_detail, exception_code)
        self.method_name = method_name
        self.exception_code = exception_code
        self.exception_detail = exception_detail
        self.exception_traceback = exception_traceback


class CoreStripeException(Exception):
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR

    def __init__(self, exception_code=None, method_name=None,
                 exception_detail="There is some issue occurred during transaction. Please contact support.",
                 exception_traceback=None):
        super().__init__(exception_detail, exception_code)
        self.method_name = method_name
        self.exception_code = exception_code
        self.exception_detail = exception_detail
        self.exception_traceback = exception_traceback



from datetime import datetime
import pytz
import stripe
from src.common.constants.constant import *
from src.exceptions.exception_constants import *

from src.exceptions.exception_handling import *



class StripeSubscription:

    @staticmethod
    def create_stripe_subscription(
            stripe_secret_key: str, customer_id: str,
            price_id: str, currency: str,
            coupon: str = None, promo_code: str = None,
            cancel_at_period_end: bool = False, default_payment_method: str = None,
            metadata: dict = None, description: str = None
    ) -> dict:
        """
        A stripe create subscription method.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account’s API keys. Stripe raises an invalid request
                                  error if you don’t include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type customer_id: string
        :param customer_id: Stripe id of the customer to subscribe.

        :type price_id: string
        :param price_id: Stripe id of the price object.

        :type currency: string
        :param currency: Three-letter ISO currency code, in lowercase.
                         Must be a <a href="https://stripe.com/docs/currencies">
                         supported currency</a>.

        :type coupon: string
        :param coupon: The ID of the coupon to apply to this subscription.
                       A coupon applied to a subscription will only affect
                       invoices created for that particular subscription.

        :type promo_code: string
        :param promo_code: The API ID of a promotion code to apply to this subscription.
                           A promotion code applied to a subscription will only affect
                           invoices created for that particular subscription.

        :type cancel_at_period_end: boolean
        :param cancel_at_period_end: Boolean indicating whether this subscription should
                                     cancel at the end of the current period.

        :type default_payment_method: string
        :param default_payment_method: ID of the default payment method for the subscription.
                                       It must belong to the customer associated with the
                                       subscription.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can be useful
                         for storing additional information about the object in a structured format.
                         Individual keys can be unset by posting an empty value to them. All keys
                         can be unset by posting an empty value to metadata.

        :type description: string
        :param description: The subscription’s description, meant to be displayable to the customer.
                            Use this field to optionally store an explanation of the subscription
                            for rendering in Stripe surfaces.

        :type: dictionary
        :returns: The newly created Subscription object, if the call succeeded.
                  If the attempted charge fails, the subscription is created in an incomplete
                  status.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(customer_id, str) \
                    or not isinstance(price_id, str) \
                    or not isinstance(currency, str) \
                    or not isinstance(coupon, (str, type(None))) \
                    or not isinstance(promo_code, (str, type(None))) \
                    or not isinstance(cancel_at_period_end, bool) \
                    or not isinstance(default_payment_method, (str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(description, (str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_CREATE_SUBSCRIPTION_METHOD)

            subscription = stripe.Subscription.create(
                customer=customer_id,
                items=[
                    {
                        "price": price_id
                    }
                ],
                metadata=metadata,
                coupon=coupon,
                cancel_at_period_end=cancel_at_period_end,
                default_payment_method=default_payment_method,
                description=description,
                payment_behavior="default_incomplete",
                currency=currency,
                promotion_code=promo_code,
                collection_method="charge_automatically",
                proration_behavior=None,
                expand=["latest_invoice.payment_intent"],
                api_key=stripe_secret_key
            )

            return {
                DATA: subscription,
                MESSAGE: "Subscription created successfully."
            }

        except stripe.error.InvalidRequestError as exception:
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST_MESSAGE,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_CREATE_SUBSCRIPTION_METHOD)

        except stripe.error.StripeError as exception:
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_CREATE_SUBSCRIPTION_METHOD)

        except CoreBadRequest as exception:
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_CREATE_SUBSCRIPTION_METHOD)

    def update_stripe_subscription(
            self, stripe_secret_key: str,
            subscription_id: str, price_id: str = None,
            coupon: str = None, promo_code: str = None,
            default_payment_method: str = None, metadata: dict = None,
            description: str = None
    ) -> dict:
        """
        A stripe update subscription method. It will update the subscription at the
        current period end.

        :type stripe_secret_key: string
        :param stripe_secret_key: Stripe authenticates your API requests using your
                                  account’s API keys. Stripe raises an invalid request
                                  error if you don’t include a key, and an authentication
                                  error if the key is incorrect or outdated.

        :type subscription_id: string
        :param subscription_id: Stripe subscription id.

        :type price_id: string
        :param price_id: Stripe price id.

        :type coupon: string
        :param coupon: The ID of the coupon to apply to this subscription.
                       A coupon applied to a subscription will only affect
                       invoices created for that particular subscription.

        :type promo_code: string
        :param promo_code: The API ID of a promotion code to apply to this subscription.
                           A promotion code applied to a subscription will only affect
                            invoices created for that particular subscription.

        :type default_payment_method: string
        :param default_payment_method: ID of the default payment method for the subscription.
                                       It must belong to the customer associated with the
                                       subscription.

        :type metadata: dictionary
        :param metadata: Set of key-value pairs that you can attach to an object. This can be useful
                         for storing additional information about the object in a structured format.
                         Individual keys can be unset by posting an empty value to them. All keys
                         can be unset by posting an empty value to metadata.

        :type description: string
        :param description: The subscription’s description, meant to be displayable to the customer.
                            Use this field to optionally store an explanation of the subscription
                            for rendering in Stripe surfaces.

        :type: dictionary
        :returns: The newly updated Subscription object, if the call succeeded.
        """
        try:
            if not isinstance(stripe_secret_key, str) \
                    or not isinstance(subscription_id, str) \
                    or not isinstance(price_id, (str, type(None))) \
                    or not isinstance(coupon, (str, type(None))) \
                    or not isinstance(promo_code, (str, type(None))) \
                    or not isinstance(default_payment_method, (str, type(None))) \
                    or not isinstance(metadata, (dict, type(None))) \
                    or not isinstance(description, (str, type(None))):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_MODIFY_SUBSCRIPTION_METHOD)

            # stripe retrieve stripe subscription
            subscription_obj = self.retrieve_stripe_subscription(
                stripe_secret_key=stripe_secret_key, subscription_id=subscription_id
            )

            # get coupon from promo-code
            if coupon is None and promo_code is not None:
                # TODO: Call from promo_code pkg
                promo_code_obj = stripe.PromotionCode.retrieve(
                    promo_code, api_key=stripe_secret_key
                )
                coupon = promo_code_obj.get("coupon").get("id")

            # get scheduler_id from subscription object or create scheduler for subscription
            if subscription_obj.get("schedule") is None:
                schedule_obj = StripeSubscriptionSchedule().create_stripe_subscription_schedule(
                    stripe_secret_key=stripe_secret_key, subscription_id=subscription_id
                )
                scheduler_id = schedule_obj.get("data").get("id")
            else:
                scheduler_id = subscription_obj.get("schedule")

            if price_id is not None or \
                    (price_id is None and coupon is not None):

                # Update subscription via subscription schedule
                StripeSubscriptionSchedule().update_stripe_subscription_schedule(
                    stripe_secret_key=stripe_secret_key,
                    schedule_id=scheduler_id,
                    current_phase_start_dt=subscription_obj.get("current_period_start"),
                    current_phase_end_dt=subscription_obj.get("current_period_end"),
                    current_price_id=subscription_obj.get("plan").get("id"),
                    new_price_id=price_id,
                    coupon=coupon,
                    default_payment_method=default_payment_method,
                    metadata=metadata,
                    description=description
                )

                # get updated subscription object
                subscription_obj = self.retrieve_stripe_subscription(
                    stripe_secret_key=stripe_secret_key,
                    subscription_id=subscription_id
                )

            else:
                # Update subscription when not having price related changes
                subscription_obj = stripe.Subscription.modify(
                    subscription_id,
                    metadata=metadata,
                    default_payment_method=default_payment_method,
                    description=description,
                    expand=["latest_invoice.payment_intent"],
                    api_key=stripe_secret_key
                )

            return {
                DATA: subscription_obj,
                MESSAGE: "Subscription changes for upgrade/downgrade are applied at the end "
                         "of your current billing cycle."
            }

        except stripe.error.InvalidRequestError as exception:
            log_error(exception, STRIPE_INVALID_REQUEST, STRIPE_MODIFY_SUBSCRIPTION_METHOD)
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_MODIFY_SUBSCRIPTION_METHOD)

        except stripe.error.StripeError as exception:
            log_error(exception, STRIPE_GENERIC, STRIPE_MODIFY_SUBSCRIPTION_METHOD)
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_MODIFY_SUBSCRIPTION_METHOD)

        except CoreBadRequest as exception:
            log_error(exception.exception_traceback if exception.exception_traceback else exception,
                      exception.exception_code, method_name=exception.method_name)
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            log_error(exception, STRIPE_GENERIC, STRIPE_MODIFY_SUBSCRIPTION_METHOD)
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_MODIFY_SUBSCRIPTION_METHOD)

    def cancel_stripe_subscription(
            self, stripe_secret_key: str, subscription_id: str
    ) -> dict:
        """
            A stripe cancel subscription method. It will cancel the subscription at the
            current period end.

            :type stripe_secret_key: string
            :param stripe_secret_key: Stripe authenticates your API requests using your
                                      account’s API keys. Stripe raises an invalid request
                                      error if you don’t include a key, and an authentication
                                      error if the key is incorrect or outdated.

            :type subscription_id: string
            :param subscription_id: Stripe subscription id.

            :type: dictionary
            :returns: The canceled Subscription object. Its subscription status will set to be
                      canceled.
        """
        try:
            if not isinstance(subscription_id, str) \
                    or not isinstance(stripe_secret_key, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_CANCEL_SUBSCRIPTION_METHOD)

            # stripe retrieve stripe subscription
            subscription_obj = self.retrieve_stripe_subscription(
                stripe_secret_key=stripe_secret_key, subscription_id=subscription_id
            )

            # convert unix time to utc timestamp
            current_period_end = pytz.utc.localize(
                datetime.fromtimestamp(subscription_obj.get("current_period_end")))

            now = pytz.utc.localize(datetime.now())

            if subscription_obj.get("schedule") is not None:

                # release stripe subscription schedule for cancel at period end
                StripeSubscriptionSchedule().release_stripe_subscription_schedule(
                    stripe_secret_key=stripe_secret_key,
                    schedule_id=subscription_obj.get("schedule")
                )

            if now < current_period_end:
                subscription = stripe.Subscription.modify(
                    subscription_id,
                    cancel_at_period_end=True,
                    api_key=stripe_secret_key
                )
                return {
                    DATA: subscription,
                    MESSAGE: "Subscription would be cancelled on end day of your billing period."
                }

            else:
                subscription = stripe.Subscription.delete(
                    subscription_id, api_key=stripe_secret_key
                )
                return {
                    DATA: subscription,
                    MESSAGE: "Subscription cancelled."
                }

        except stripe.error.InvalidRequestError as exception:
            log_error(exception, STRIPE_INVALID_REQUEST, STRIPE_CANCEL_SUBSCRIPTION_METHOD)
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_CANCEL_SUBSCRIPTION_METHOD)

        except stripe.error.StripeError as exception:
            log_error(exception, STRIPE_GENERIC, STRIPE_CANCEL_SUBSCRIPTION_METHOD)
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_CANCEL_SUBSCRIPTION_METHOD)

        except CoreBadRequest as exception:
            log_error(exception.exception_traceback if exception.exception_traceback else exception,
                      exception.exception_code, method_name=exception.method_name)
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            log_error(exception, STRIPE_GENERIC, STRIPE_CANCEL_SUBSCRIPTION_METHOD)
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_CANCEL_SUBSCRIPTION_METHOD)

    @staticmethod
    def retrieve_stripe_subscription(
            stripe_secret_key: str, subscription_id: str
    ) -> dict:
        """
            A stripe retrieve subscription method.

            :type stripe_secret_key: string
            :param stripe_secret_key: Stripe authenticates your API requests using your
                                      account’s API keys. Stripe raises an invalid request
                                      error if you don’t include a key, and an authentication
                                      error if the key is incorrect or outdated.

            :type subscription_id: string
            :param subscription_id: Stripe subscription id

            :type: dictionary
            :returns: Returns the subscription object.
        """

        try:
            if not isinstance(subscription_id, str) \
                    or not isinstance(stripe_secret_key, str):
                raise CoreBadRequest(exception_detail=BAD_REQUEST_MESSAGE,
                                     exception_code=STRIPE_BAD_REQUEST,
                                     method_name=STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)

            subscription = stripe.Subscription.retrieve(
                subscription_id, api_key=stripe_secret_key
            )

            return {
                DATA: subscription,
                MESSAGE: "Subscription retrieved successfully."
            }

        except stripe.error.InvalidRequestError as exception:
            log_error(exception, STRIPE_INVALID_REQUEST, STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)
            raise CoreBadRequest(exception_detail=STRIPE_INVALID_REQUEST,
                                 exception_code=STRIPE_INVALID_REQUEST,
                                 exception_traceback=exception,
                                 method_name=STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)

        except stripe.error.StripeError as exception:
            log_error(exception, STRIPE_GENERIC, STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)
            raise CoreStripeException(exception_code=STRIPE_GENERIC,
                                      exception_traceback=exception,
                                      method_name=STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)

        except CoreBadRequest as exception:
            log_error(exception.exception_traceback if exception.exception_traceback else exception,
                      exception.exception_code, method_name=exception.method_name)
            raise CoreBadRequest(exception_detail=exception.exception_detail,
                                 exception_code=exception.exception_code,
                                 method_name=exception.method_name)

        except Exception as exception:
            log_error(exception, STRIPE_GENERIC, STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)
            raise CoreGenericException(exception_code=STRIPE_GENERIC,
                                       exception_traceback=exception,
                                       method_name=STRIPE_RETRIEVE_SUBSCRIPTION_METHOD)



# response message
BAD_REQUEST_MESSAGE = "All the required information is not in a proper format."
USER_NOT_EXIST = "User does not exist."
GENERIC_MESSAGE = "There is some issue. Please contact support."

# General constants
MESSAGE = "message"
DATA = "data"
CARD = "card"
HUNDRED = "100"